import asyncio
import logging
from dataclasses import dataclass, field, fields, replace
from inspect import iscoroutinefunction, signature
from typing import (
    Any,
    AsyncIterator,
    Collection,
    Dict,
    List,
    Type,
    TypeVar,
    Union,
    cast,
)

from .exception import ModelNotFoundException
from .fields import deserialize, is_cascade, is_transient, serialize, update_field
from .session import connection, transaction

_logger = logging.getLogger(__name__)


class ModelDataclassType(type):
    def __new__(cls, name, bases, dict):
        for field_name, field_type in dict.get("__annotations__", {}).items():
            update_field(field_name, field_type, dict)
        model_class = dataclass(
            super().__new__(cls, name, bases, dict), unsafe_hash=True
        )
        setattr(
            model_class,
            "NotFoundException",
            type("NotFoundException", (ModelNotFoundException,), {}),
        )
        return model_class


T = TypeVar("T", bound="Model")


class Model(metaclass=ModelDataclassType):
    id: int = field(init=True, repr=False, compare=False)

    @classmethod
    def prefix(cls) -> str:
        return f"{cls.__name__.lower()}"

    @classmethod
    async def get(cls: Type[T], id: Union[int, str]) -> T:
        async with connection() as conn:
            db_item = cast(Dict[str, Any], await conn.hgetall(f"{cls.prefix()}:{id}"))

        if not db_item:
            raise cls.NotFoundException(f"{id} not found")  # type: ignore

        model_fields = [f for f in fields(cls) if not is_transient(f)]
        deserialized = await asyncio.gather(
            *[deserialize(f, db_item.get(f.name)) for f in model_fields]
        )

        return cls.from_dict(
            {f.name: value for f, value in zip(model_fields, deserialized)},
            strict=False,
        )

    @classmethod
    def from_dict(cls: Type[T], model: Dict[str, Any], strict: bool = True) -> T:
        parameters = signature(cls).parameters
        return (
            cls(**{k: v for k, v in model.items() if k in parameters})
            if strict
            else cls(**model)
        )

    @classmethod
    async def scan(cls: Type[T], **kwargs) -> AsyncIterator[T]:
        async with connection() as conn:
            found = set()
            async for key in conn.isscan(cls.prefix(), **kwargs):
                if key not in found:
                    value = cast(T, await cls.get(key))
                    if value:
                        yield value
                        found.add(key)
                    else:
                        _logger.warning(f"{cls.__name__} Key: {key} orphaned")

    @classmethod
    async def all(cls: Type[T]) -> List[T]:
        async with connection() as conn:
            keys = await conn.smembers(cls.prefix())
            return await asyncio.gather(*[cls.get(key) for key in keys])

    @staticmethod
    async def flush():
        async with connection() as conn:
            await conn.flushdb()

    @classmethod
    async def count(cls) -> int:
        async with connection() as conn:
            return await conn.scard(cls.prefix())

    @classmethod
    async def delete_all(cls: Type):
        key_prefix = cls.prefix()
        async with connection() as conn:
            keys = await conn.keys(f"{key_prefix}:*")
            await conn.delete(key_prefix, *keys)

    @classmethod
    async def persisted(cls: Type, id: int) -> bool:
        async with connection() as conn:
            return await conn.exists(f"{cls.prefix()}:{id}")

    @property
    def db_id(self) -> str:
        return f"{self.prefix()}:{self.id}"

    async def save(self, optimistic=False):
        async with transaction() as tr:
            model_dict = await self._serialized_model(optimistic)
            tr.hmset_dict(self.db_id, model_dict)
            tr.sadd(self.prefix(), self.id)

    async def update(self, optimistic=False, **changes) -> T:
        async with transaction() as tr:
            model_dict = await self._serialized_model(optimistic, **changes)
            for key, value in model_dict.items():
                tr.hset(self.db_id, key, value)
            return replace(self, **changes)

    async def _serialized_model(self, optimistic, **changes) -> Dict[str, Any]:
        async with connection() as conn:
            if optimistic:
                conn.watch(self.db_id)
            model_fields = {
                f: changes.get(f.name, getattr(self, f.name))
                for f in fields(self)
                if not is_transient(f)
                and hasattr(self, f.name)
                and (not changes or f.name in changes)
            }

            serialized = await asyncio.gather(
                *[
                    serialize(field, f"{self.db_id}:{field.name}", value)
                    for field, value in model_fields.items()
                ]
            )
            serialized_fields = list(zip(model_fields.keys(), serialized))

            await asyncio.gather(
                *[
                    value.save(optimistic=optimistic)
                    for field, value in serialized_fields
                    if iscoroutinefunction(getattr(value, "save", None))
                    and (is_cascade(field) or isinstance(value, Collection))
                ]
            )
            return {
                field.name: getattr(value, "id", f"{self.db_id}:{field.name}")
                if iscoroutinefunction(getattr(value, "save", None))
                else value
                for field, value in serialized_fields
                if value is not None
                or iscoroutinefunction(getattr(value, "save", None))
            }

    async def delete(self):
        key = self.db_id
        async with connection() as conn:
            keys = await conn.keys(f"{key}:*")
            async with transaction() as tr:
                tr.delete(*keys, key)
                tr.srem(self.prefix(), key)

    async def exists(self) -> bool:
        async with connection() as conn:
            return await conn.exists(self.db_id)

    async def refresh(self: T) -> T:
        refreshed = await type(self).get(self.id)
        return cast(T, refreshed)



import asyncio
import os
from typing import List
from unittest import TestCase, skipUnless

import pytest
from rom import Model, field
from rom.session import connection


class Bar(Model):
    field1: int
    field2: str
    field3: List[int]


class Foo(Model):
    bars: List[Bar] = field(eager=True, cascade=True, default_factory=list)


@skipUnless(os.environ.get("CI"), "Redis benchmark CI test only")
class Benchmark(TestCase):
    items = 100

    @pytest.fixture(autouse=True)
    def setupBenchmark(self, benchmark):
        self.benchmark = benchmark

    def run_coro(self, coro_factory):
        return asyncio.run(coro_factory())

    async def delete_all(self):
        await Bar.delete_all()
        await Foo.delete_all()

    def setUp(self):
        self.bar = Bar(1, 123, "value", [1, 2, 3])
        asyncio.run(self.bar.save())

    def tearDown(self):
        asyncio.run(self.delete_all())

    def test_save(self):
        async def save():
            async with connection():
                for _ in range(self.items):
                    await self.bar.save()

        self.benchmark(self.run_coro, save)

    def test_get(self):
        async def get():
            async with connection():
                for _ in range(self.items):
                    await Bar.get(1)

        self.benchmark(self.run_coro, get)

    def test_get_eager_list(self):
        foo = Foo(1)
        for i in range(self.items):
            foo.bars.append(Bar(i, 123, "value", [1, 2, 3]))
        asyncio.run(foo.save())

        async def get():
            async with connection():
                return await Foo.get(1)

        result = self.benchmark(self.run_coro, get)
        assert self.items == len(result.bars)

    def test_get_all(self):
        for i in range(self.items):
            asyncio.run(Bar(i, 123, "value", [1, 2, 3]).save())

        result = self.benchmark(self.run_coro, Bar.all)
        assert self.items == len(result)

    def test_scan_all(self):
        for i in range(self.items):
            asyncio.run(Bar(i, 123, "value", [1, 2, 3]).save())

        async def scan():
            result = []
            async for item in Bar.scan():
                result.append(item)
            return result

        result = self.benchmark(self.run_coro, scan)
        assert self.items == len(result)

    def test_cascade_save(self):
        foo = Foo(1)
        for i in range(self.items):
            foo.bars.append(Bar(i, 123, "value", [1, 2, 3]))
        self.benchmark(self.run_coro, foo.save)



import os
from typing import List, Optional, Set

from asynctest import TestCase, skipUnless
from rom import Model, field
from rom.session import redis_pool


class Bar(Model):
    field1: int
    field2: str
    field3: List[int] = field(eager=True, hash=False)


class Foo(Model):
    eager_bars: List[Bar] = field(eager=True, hash=False)
    f1: Optional[str] = None
    lazy_bars: Set[Bar] = field(compare=False, cascade=True)


class FooBar(Model):
    foos: Set[Foo] = field(eager=True, cascade=True)


@skipUnless(os.environ.get("CI"), "Redis CI test only")
class RedisIntegrationTestCase(TestCase):
    async def setUp(self):
        self.bar = Bar(1, 123, "value", [1, 2, 3])

    async def tearDown(self):
        await Foo.delete_all()
        await Bar.delete_all()
        await FooBar.delete_all()

    async def test_save(self):
        await self.bar.save()

        async with redis_pool() as redis:
            field1 = await redis.hget("bar:1", "field1")
            field2 = await redis.hget("bar:1", "field2")
            field3 = await redis.hget("bar:1", "field3")
            field3_value = await redis.lrange("bar:1:field3", 0, -1)

        assert "123" == field1
        assert "value" == field2
        assert "bar:1:field3" == field3
        assert ["1", "2", "3"] == field3_value

    async def test_get(self):
        await self.bar.save()
        bar = await Bar.get(1)
        assert self.bar == bar

    async def test_get_with_references(self):
        await self.bar.save()
        foo = Foo(123, [self.bar], None, set([self.bar]))
        await foo.save()
        gotten_foo = await Foo.get(123)
        assert foo == gotten_foo
        await gotten_foo.lazy_bars.load()
        for bar in gotten_foo.lazy_bars:
            assert bar in foo.lazy_bars
        assert len(foo.lazy_bars) == len(gotten_foo.lazy_bars)

    async def _test_collection_references(self, test_cascade=False):
        await self.bar.save()
        foo = Foo(123, [self.bar], None, set([self.bar]))
        if not test_cascade:
            await foo.save()
        foobar = FooBar(321, set([foo]))
        await foobar.save()

        gotten_foobar = await FooBar.get(321)
        assert foobar == gotten_foobar
        assert set([foo]) == gotten_foobar.foos
        for gotten_foo in gotten_foobar.foos:
            assert 1 == len(gotten_foo.eager_bars)
            await gotten_foo.lazy_bars.load()
            for bar in gotten_foo.lazy_bars:
                assert bar in foo.lazy_bars

    async def test_collections(self):
        await self._test_collection_references()

    async def test_collection_cascades_references(self):
        await self._test_collection_references(test_cascade=True)

    async def test_update_collection_references(self):
        await self.bar.save()
        foo = Foo(123, [self.bar], None, set([self.bar]))
        foobar = FooBar(321, set([foo]))
        await foobar.save()
        refreshed = await foobar.refresh()
        foo2 = Foo(222, [], None, set([]))
        refreshed.foos.add(foo2)
        await refreshed.save()

        gotten_foobar = await FooBar.get(321)
        assert refreshed == gotten_foobar
        assert set([foo, foo2]) == gotten_foobar.foos

    async def test_update(self):
        await self.bar.save()
        await self.bar.update(field2="updated")
        async with redis_pool() as redis:
            field2 = await redis.hget("bar:1", "field2")
        assert "updated" == field2
        bar = await Bar.get(1)
        assert "updated" == bar.field2

    async def test_update_reference(self):
        await self.bar.save()
        foo = Foo(123, [self.bar], None, set([self.bar]))
        await foo.save()

        bar2 = Bar(2, 123, "otherbar", [1, 2, 3, 4])
        await bar2.save()

        foo = await foo.update(lazy_bars=set([bar2]))
        async with redis_pool() as redis:
            lazy_bars = await redis.smembers("foo:123:lazy_bars")
            assert ["2"] == lazy_bars

        foo = await foo.update(eager_bars=[bar2])
        async with redis_pool() as redis:
            eager_bars = await redis.lrange("foo:123:eager_bars", 0, -1)
            assert ["2"] == eager_bars

        gotten_foo = await Foo.get(123)
        assert foo == gotten_foo

    async def test_save_again_overrides_previous(self):
        await self.bar.save()
        bar = await Bar.get(1)
        bar.field2 = "updated"
        await bar.save()
        async with redis_pool() as redis:
            field2 = await redis.hget("bar:1", "field2")
        assert "updated" == field2

    async def test_delete(self):
        await self.bar.save()
        async with redis_pool() as redis:
            assert await redis.exists("bar:1")
            await self.bar.delete()
            assert not await redis.exists("bar:1")

    async def test_delete_all(self):
        await self.bar.save()
        async with redis_pool() as redis:
            await Bar.delete_all()
            assert not await redis.keys("bar*")

    async def test_lazy_collection_cascade(self):
        foo = Foo(123, [self.bar], None, set([self.bar]))
        await foo.save()
        foo = await Foo.get(123)
        other_bar = Bar(2, 124, "value2", [])
        foo.lazy_bars.add(other_bar)
        await foo.save()
        gotten_foo = await Foo.get(123)
        assert foo == gotten_foo
        await gotten_foo.lazy_bars.load()
        await foo.lazy_bars.load()
        assert 2 == len(foo.lazy_bars) == len(gotten_foo.lazy_bars)



from aioredis import Redis
from asynctest import CoroutineMock, MagicMock, TestCase
from asynctest.mock import patch
from rom import Model
from rom.exception import ModelNotFoundException


class ForTesting(Model):
    f1: int


class ModelTestCase(TestCase):
    def setUpClass():
        ModelTestCase.redis_await = Redis.__await__
        # Redis & ContextRedis are awaitable, asynctest will try to yield from them
        delattr(Redis, "__await__")

    def tearDownClass():
        setattr(Redis, "__await__", ModelTestCase.redis_await)

    async def setUp(self):
        self.mock_redis_transaction = MagicMock(spec=Redis)
        self.mock_redis_transaction.execute.side_effect = CoroutineMock(spec=Redis)
        self.mock_redis_client = MagicMock(spec=Redis)
        self.mock_redis_client.multi_exec.return_value = self.mock_redis_transaction
        self.mock_redis_client.smembers.side_effect = CoroutineMock()
        self.transaction_patcher = patch("rom.model.transaction")
        self.transaction_mock = self.transaction_patcher.start()
        self.transaction_mock.return_value.__aenter__.return_value = (
            self.mock_redis_transaction
        )
        self.connection_patcher = patch("rom.model.connection")
        self.connection_mock = self.connection_patcher.start()
        self.connection_mock.return_value.__aenter__.return_value = (
            self.mock_redis_client
        )

    async def tearDown(self):
        patch.stopall()

    async def test_save(self):
        await ForTesting(123, 123).save()
        self.mock_redis_transaction.hmset_dict.assert_called_with(
            "fortesting:123", {"id": "123", "f1": "123"}
        )
        self.mock_redis_transaction.sadd.assert_called_with("fortesting", 123)

    async def test_update(self):
        await ForTesting(123, 123).update(f1=987)
        self.mock_redis_transaction.hset.assert_called_once_with(
            "fortesting:123", "f1", "987"
        )
        self.mock_redis_transaction.sadd.assert_not_called()

    async def test_get(self):
        self.mock_redis_client.hgetall.side_effect = CoroutineMock(
            return_value={"id": "123", "f1": "123"}
        )
        value = await ForTesting.get(123)
        assert 123 == value.id
        assert 123 == value.f1
        assert ForTesting(123, 123) == value

    async def test_failed_get(self):
        self.mock_redis_client.hgetall.side_effect = CoroutineMock(return_value=None)
        await self.assertAsyncRaises(ModelNotFoundException, ForTesting.get(123))

    async def test_scan(self):
        self.mock_redis_client.hgetall.side_effect = CoroutineMock(
            side_effect=[{"id": "123", "f1": "123"}, {"id": "124", "f1": "123"}]
        )
        keys = MagicMock()
        keys.__aiter__.return_value = ["123", "124"]
        self.mock_redis_client.isscan.return_value = keys
        items = 0
        async for obj in ForTesting.scan():
            assert 123 == obj.f1
            assert isinstance(obj.id, int)
            items += 1
        assert 2 == items

    async def test_all(self):
        self.mock_redis_client.smembers.side_effect.return_value = ["123", "124"]
        self.mock_redis_client.hgetall.side_effect = CoroutineMock(
            side_effect=[{"id": "123", "f1": "123"}, {"id": "124", "f1": "123"}]
        )
        items = 0
        for obj in await ForTesting.all():
            assert 123 == obj.f1
            assert isinstance(obj.id, int)
            items += 1
        assert 2 == items

    async def test_count(self):
        self.mock_redis_client.scard.side_effect = CoroutineMock(return_value=10)
        assert 10 == await ForTesting.count()

    async def test_delete(self):
        self.mock_redis_client.keys.side_effect = CoroutineMock(
            return_value=["fortesting:123:reference"]
        )
        await ForTesting(123, 987).delete()
        self.mock_redis_transaction.delete.assert_called_with(
            "fortesting:123:reference", "fortesting:123"
        )
        self.mock_redis_transaction.srem.assert_called_with(
            "fortesting", "fortesting:123"
        )

    async def test_delete_all(self):
        self.mock_redis_client.keys.side_effect = CoroutineMock(
            return_value=["fortesting:1"]
        )
        delete = CoroutineMock()
        self.mock_redis_client.delete.side_effect = delete
        await ForTesting.delete_all()
        delete.assert_called_with("fortesting", "fortesting:1")

    async def test_flush(self):
        self.mock_redis_client.flushdb.side_effect = CoroutineMock()
        await ForTesting.flush()
        self.mock_redis_client.flushdb.assert_called_once()

    async def test_exists(self):
        self.mock_redis_client.exists.side_effect = CoroutineMock(
            side_effect=[True, False, True, False]
        )
        assert await ForTesting.persisted(1)
        assert not await ForTesting.persisted(1)
        assert await ForTesting(1, "123").exists()
        assert not await ForTesting(1, "123").exists()

    async def test_refresh(self):
        self.mock_redis_client.hgetall.side_effect = CoroutineMock(
            return_value={"id": "123", "f1": "124"}
        )
        value = await ForTesting(123, 123).refresh()
        assert 123 == value.id
        assert 124 == value.f1
        assert ForTesting(123, 124) == value



import json
from asyncio.coroutines import iscoroutine
from collections.abc import MutableSequence
from dataclasses import Field
from dataclasses import field as dc_field
from dataclasses import is_dataclass
from enum import Enum, auto
from functools import partial
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from typing_inspect import get_args, get_origin, is_optional_type

from .attributes import RedisList, RedisModelList, RedisModelSet, RedisSet

if TYPE_CHECKING:
    from .model import Model

T = TypeVar("T", bound="Model")


class FieldMetadata(str, Enum):
    TRANSIENT = auto()
    EAGER = auto()
    CASCADE = auto()
    OPTIONAL = auto()
    DESERIALIZER = auto()
    SERIALIZER = auto()


def is_transient(field: Field):
    return field.metadata.get(FieldMetadata.TRANSIENT, False)


def is_eager(field: Field):
    return field.metadata.get(FieldMetadata.EAGER, False)


def is_cascade(field: Field):
    return field.metadata.get(FieldMetadata.CASCADE, False)


def is_optional(field: Field):
    return field.metadata.get(FieldMetadata.OPTIONAL, False)


def is_model(model: Union[Type, object]):
    return is_dataclass(model) and hasattr(model, "prefix")


def field(
    transient: bool = False, cascade: bool = False, eager: bool = False, **kwargs
) -> Any:
    metadata = kwargs.pop("metadata", {})
    metadata[FieldMetadata.TRANSIENT] = transient
    metadata[FieldMetadata.CASCADE] = cascade
    metadata[FieldMetadata.EAGER] = eager
    return dc_field(metadata=metadata, **kwargs)


def deserialize_reference(model_class: Type[T]):
    async def deserializer(
        key: Union[int, str],
    ) -> Optional[T]:
        if key is None:
            return None
        return cast(T, await model_class.get(key))

    return deserializer


async def deserialize(field: Field, value):
    if value is None and is_optional(field):
        return None
    val = field.metadata.get(FieldMetadata.DESERIALIZER)(value)
    if iscoroutine(val):
        val = await val
    return val


async def serialize(field: Field, key: str, value):
    val = field.metadata.get(FieldMetadata.SERIALIZER)(key, value)
    if iscoroutine(val):
        val = await val
    return val


def default_serializer(key, value):
    return json.dumps(value)


def passthrough_serializer(key, value):
    return value


def passthrough_deserializer(value):
    return value


def update_field(name, field_type, fields: Dict[str, Field]):
    field = fields.get(name, dc_field()) or dc_field()
    origin = get_origin(field_type) or field_type
    type_args = get_args(field_type)
    metadata = dict(getattr(field, "metadata", {}))
    eager = is_eager(field)
    cascade = is_cascade(field)
    optional = False
    deserializer = json.loads
    serializer = default_serializer

    # Unwrap optional type
    if is_optional_type(field_type):
        optional = True
        field_type = type_args[0]
        type_args = get_args(field_type)

    if isinstance(field_type, type) and issubclass(field_type, str):
        deserializer = passthrough_deserializer
        serializer = passthrough_serializer
    elif is_model(field_type):
        deserializer = deserialize_reference(field_type)
        serializer = passthrough_serializer
    elif issubclass(origin, AbstractSet):
        if is_model(type_args[0]):
            deserializer = partial(
                RedisModelSet.from_key,
                field.default_factory,
                model_class=type_args[0],
                eager=eager,
                cascade=cascade,
            )
            serializer = partial(
                RedisModelSet.serialize, model_class=type_args[0], cascade=cascade
            )
        else:
            deserializer = partial(
                RedisSet.from_key, field.default_factory, eager=eager
            )
            serializer = RedisSet
    elif issubclass(origin, MutableSequence):
        if is_model(type_args[0]):
            deserializer = partial(
                RedisModelList.from_key,
                field.default_factory,
                model_class=type_args[0],
                eager=eager,
                cascade=cascade,
            )
            serializer = partial(
                RedisModelList.serialize, model_class=type_args[0], cascade=cascade
            )
        else:
            deserializer = partial(
                RedisList.from_key, field.default_factory, eager=eager
            )
            serializer = RedisList

    metadata[FieldMetadata.DESERIALIZER] = deserializer
    metadata[FieldMetadata.SERIALIZER] = serializer
    metadata[FieldMetadata.OPTIONAL] = optional
    field.metadata = MappingProxyType(metadata)
    fields[name] = field



import asyncio
import json
from abc import ABCMeta, abstractmethod
from asyncio.coroutines import iscoroutine
from dataclasses import MISSING
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Collection,
    Dict,
    Generic,
    Iterator,
    MutableSequence,
    MutableSet,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

if TYPE_CHECKING:
    from .model import Model

from .exception import ModelNotLoadedException
from .session import connection, transaction

T = TypeVar("T", bound="Model")


class RedisCollection(Collection, metaclass=ABCMeta):
    def __init__(self, key: str, values: Sequence):
        self._key = key
        self.values = values

    @classmethod
    async def from_key(cls, default_factory, key: str, eager: bool = False, **kwargs):
        values = await cls.get_values_for_key(key)
        if values is None:
            return default_factory() if default_factory != MISSING else None

        collection = cls(
            key, type(values)([json.loads(value) for value in values]), **kwargs
        )
        if eager:
            await collection.load()
            return collection.values
        return collection

    @abstractmethod
    async def save(self, optimistic: bool = False):
        pass

    async def load(self):
        pass

    @classmethod
    @abstractmethod
    async def get_values_for_key(cls, key: str) -> Sequence:
        pass

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, value):
        self._values = value

    def __iter__(self) -> Iterator:
        return filter(None, iter(self.values))

    def __len__(self) -> int:
        return len(self.values)

    def __contains__(self, x) -> bool:
        return x in self.values

    def __aiter__(self):
        self._iter = iter(self.values)
        return self

    async def __anext__(self):
        try:
            val = next(self._iter)
            if iscoroutine(val):
                return await val
            return val
        except StopIteration:
            raise StopAsyncIteration


class RedisSet(MutableSet, RedisCollection):
    @classmethod
    async def get_values_for_key(cls, id: int) -> AbstractSet:
        async with connection() as conn:
            return set(await conn.smembers(id))

    async def save(self, optimistic: bool = False):
        async with connection() as conn:
            if optimistic:
                conn.watch(self._key)
            async with transaction() as tr:
                tr.delete(self._key)
                values = super().values
                if values:
                    tr.sadd(self._key, *values)

    def add(self, value):
        super().values.add(value)

    def discard(self, value):
        super().values.discard(value)


class RedisList(MutableSequence, RedisCollection):
    @classmethod
    async def get_values_for_key(cls, id: int) -> Sequence:
        async with connection() as redis:
            return await redis.lrange(id, 0, -1)

    async def save(self, optimistic: bool = False):
        async with connection() as conn:
            if optimistic:
                conn.watch(self._key)
            async with transaction() as tr:
                tr.delete(self._key)
                values = super().values
                if values:
                    tr.rpush(self._key, *values)

    def insert(self, index: int, value: T):
        super().values.insert(index, value)

    def __getitem__(self, i: Union[int, slice]) -> Union[Any, Sequence[Any]]:
        return super().values.__getitem__(i)

    def __setitem__(self, i: Union[int, slice], o: Union[Any, Sequence[Any]]):
        super().values.__setitem__(i, o)

    def __delitem__(self, i: int):
        super().values.__delitem__(i)


class ModelCollection(RedisCollection, Generic[T], metaclass=ABCMeta):
    def __init__(
        self, key: str, keys: Sequence[int], model_class: Type[T], cascade: bool
    ):
        self._cache: Dict[int, T] = {}
        self._collection_type = type(keys)
        self._model_class = model_class
        self._cascade = cascade
        super().__init__(key, keys)

    @classmethod
    def serialize(
        cls, key: str, keys: Sequence, model_class: Type[T] = None, cascade=False
    ):
        return (
            keys
            if isinstance(keys, ModelCollection)
            else cls(key, keys, model_class, cascade)
        )

    async def save(self, optimistic: bool = False):
        async with transaction():
            await super().save(optimistic=optimistic)
            if self._cascade:
                for v in self.values:
                    await v.save(optimistic=optimistic)

    async def load(self):
        async with connection():
            await asyncio.gather(*[self._get_item(key) for key in super().values])

    @property
    def values(self):
        return self._collection_type(self._cache.values())

    @values.setter
    def values(self, values):
        self._cache = {v.id: v for v in values if hasattr(v, "id")}
        keys = self._collection_type(
            self._cache.keys()
            if isinstance(values, ModelCollection)
            else [getattr(v, "id", v) for v in values]
        )
        super(ModelCollection, type(self)).values.fset(self, keys)

    async def _get_item(self, key: int) -> Optional[T]:
        if key in self._cache:
            return self._cache[key]
        value = cast(T, await self._model_class.get(key))
        self._cache[key] = value
        return value

    def __aiter__(self):
        self._iter = map(self._get_item, super().values)
        return self


class RedisModelSet(ModelCollection[T], RedisSet):
    def add(self, value: T) -> None:
        key = value.id
        super().add(key)
        self._cache[key] = value

    def discard(self, value: T) -> None:
        key = value.id
        super().discard(key)
        self._cache.pop(key)


class RedisModelList(ModelCollection[T], RedisList):
    def _get_cached_item(self, key: int):
        try:
            return self._cache[key]
        except KeyError:
            raise ModelNotLoadedException(
                f"{self._model_class.__name__} object with id {key} was not loaded"
            )

    def insert(self, index: int, value: T) -> None:
        key = value.id
        super().insert(index, key)
        self._cache[key] = value

    def __getitem__(self, i: Union[int, slice]) -> Union[T, Sequence[T]]:
        if isinstance(i, slice):
            return [
                self._get_cached_item(self.values[index])
                for index in range(*i.indices(len(self)))
            ]
        else:
            return self._get_cached_item(self.values[i])

    def __setitem__(self, i: Union[int, slice], o: Union[T, Sequence[T]]) -> None:
        if isinstance(i, slice) and isinstance(o, Sequence):
            for obj_index, index in enumerate(range(*i.indices(len(self)))):
                self.insert(index, o[obj_index])
        elif not isinstance(i, slice) and not isinstance(o, Sequence):
            self.insert(i, o)
        else:
            raise ValueError

    def __delitem__(self, i: int) -> None:
        key = self.values[i]
        super().__delitem__(i)
        del self._cache[key]
